==> ised-2.2.2.tar.bz2: ised-2.2.2/src/methods.cpp <==
/** ised - generating integer and floating point sequences
    Copyright (C) 2009-2010  Simon Copar

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
**/

#include "methods.h"

using namespace std;

static void bairstow_getp(vector<ieee>& poly, ieee u, ieee v, ieee& c, ieee& d, ieee& g, ieee& h){
  int n=poly.size()-1;
  ieee b1=0;
  ieee b2=0;
  ieee f1=0;
  ieee f2=0;
  for(int i=n-2;i>=0;i--){
    ieee tmpb=poly[i+2]-u*b1-v*b2;
    ieee tmpf=b2-u*f1-v*f2;
    b2=b1;
    b1=tmpb;
    f2=f1;
    f1=tmpf;
  }
  c=poly[1]-u*b1-v*b2;
  d=poly[0]-v*b1;
  g=b2-u*f1-v*f2;
  h=b1-v*f1;
}

static int bairstow_iter(vector<ieee>& poly, ieee& u, ieee& v, ieee& error,int limit){
  //newton iteration
  ieee c,d,g,h;
  ieee det;

  int count=0;
  while(count<limit){
    bairstow_getp(poly,u,v,c,d,g,h);
    det=v*g*g+h*h-h*u*g;
    if(fabs(det)<1e-16){
      if((fabs(c)<1e-15)&&(fabs(d)<1e-15))break;//we hit the roots
      u+=AA/10;
      v+=PI/10;
      #if DEBUG>0
      fprintf(stderr,"bairstow solver: zero determinant occurred, retrying with u=%g, v=%g\n",(double)u,(double)v);
      #endif
      count++;
      continue;
    }
    det=1/det;
    ieee du=det*(-h*c+g*d);
    ieee dv=det*(-g*v*c+(g*u-h)*d);
    u-=du;
    v-=dv;
    error=sqrt(du*du+dv*dv);
    #if DEBUG>0
    fprintf(stderr,"bairstow solver: iteration step %d; error=%g\n",count,(double)error);
    #endif
    if(error<1e-15)break;
    count++;
  }
  return count;
}

static int bairstow_step(vector<ieee>& poly, vector<ieee>& lower, array& out){
  int n=poly.size()-1;
  #if DEBUG>0
  fprintf(stderr,"bairstow solver: starting step; n=%d\n",n);
  #endif
  //initial approximation
  ieee u=poly[n-1]/poly[n];
  ieee v=poly[n-2]/poly[n];
  #if DEBUG>0
  fprintf(stderr,"bairstow solver: initial approximation; u=%g, v=%g\n",(double)u,(double)v);
  #endif
  ieee error;
  int limit=n*24;
  int count=bairstow_iter(poly, u, v, error,limit);

  if((count==limit)&&(error>1e-9)){
    u=0;
    v=-pow(fabs(poly[0]/poly[n]),2./n);
    #if DEBUG>0
    fprintf(stderr,"bairstow solver: failed once, trying different approach; u=%g, v=%g\n",(double)u,(double)v);
    #endif
    count=bairstow_iter(poly, u, v, error,limit);
  }

  if(count==limit){
    if(error<1e-13){
    }else
    if(error<1e-11){
      fprintf(stderr,"bairstow solver: low accuracy (%g), possibly multiple root\n",(double)error);
    }else if(error<1e-7){
      fprintf(stderr,"bairstow solver: seriously low accuracy (%g), possibly multiple root\n",(double)error);
    }else{
      fprintf(stderr,"bairstow solver: failed to converge in %d steps\n",limit);
      return 1;
    }
  }
  //divide to obtain lower-order polynomial
  lower.resize(n-1);
  ieee b1=0;
  ieee b2=0;
  for(int i=n-2;i>=0;i--){
    ieee tmpb=poly[i+2]-u*b1-v*b2;
    lower[i]=tmpb;
    b2=b1;
    b1=tmpb;
  }
    #if DEBUG>0
    fprintf(stderr,"bairstow solver: successfully divided\n");
    #endif
  //found roots
  ieee disc=u*u-4*v;
  if(disc<-1e-14*u*u)return 0;
  if(disc<0)disc=0;
  out.push_back(value::floating(0.5*(-u+sqrt(disc))));
  out.push_back(value::floating(0.5*(-u-sqrt(disc))));
  return 0;
}

int bairstow(array& poly, array& out){
  int n2=poly.size();
  int n1=0;
  //trim zeroes
  for(int i=n2-1;i>=0;i--){
    if(poly[i].d==0){
      n2=i;
    }else break;
  }
  //front zeroes
  for(int i=0;i<n2;i++){
    if(poly[i].d==0){
      out.push_back(value::floating(0.));
      n1=i+1;
    }else{
      break;
    }
  }
  if(n2-n1>10){
    fprintf(stderr,"bairstow solver: warning, polynomial order is high (%d)\n",n2-n1);
  }
  switch(n2-n1){
    case 0:return 0;//all reals are solutions
    case 1:{//no zeroes
      return 0;
    }
    case 2:{
      out.push_back(value::floating(-poly[n1].d/poly[n1+1].d));
      return 0;
    }
    case 3:{
      ieee u=poly[n1+1].d/poly[n1+2].d;
      ieee v=poly[n1].d/poly[n1+2].d;
      ieee disc=u*u-4*v;
      if(disc<-1e-14*u*u)return 0;
      if(disc<0)disc=0;
      out.push_back(value::floating(0.5*(-u+sqrt(disc))));
      out.push_back(value::floating(0.5*(-u-sqrt(disc))));
      return 0;
    }
    default:{
       vector<ieee> p;
       for(int i=n1;i<n2;i++){
         p.push_back(poly[i].d/poly[n2-1].d);
       }
       vector<ieee> p2;
       while(p.size()>3){
         if(bairstow_step(p,p2,out)){
           return 1;
         }else{
           p=p2;
         }
       }
       //now we got polynomial of order 1 or 2:
       if(p.size()==2){
         out.push_back(value::floating(-p[0]/p[1]));
       }else{
         ieee u=p[1]/p[2];
         ieee v=p[0]/p[2];
         ieee disc=u*u-4*v;
         if(disc<-1e-14*u*u)return 0;
         if(disc<0)disc=0;
         out.push_back(value::floating(0.5*(-u+sqrt(disc))));
         out.push_back(value::floating(0.5*(-u-sqrt(disc))));
       }
       return 0;
    }
  }
}

==> ised-2.2.3.tar.bz2: ised-2.2.3/src/methods.cpp <==
/** ised - generating integer and floating point sequences
    Copyright (C) 2009-2011  Simon Copar

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
**/

#include "methods.h"

using namespace std;

static void bairstow_getp(vector<ieee>& poly, ieee u, ieee v, ieee& c, ieee& d, ieee& g, ieee& h){
  int n=poly.size()-1;
  ieee b1=0;
  ieee b2=0;
  ieee f1=0;
  ieee f2=0;
  for(int i=n-2;i>=0;i--){
    ieee tmpb=poly[i+2]-u*b1-v*b2;
    ieee tmpf=b2-u*f1-v*f2;
    b2=b1;
    b1=tmpb;
    f2=f1;
    f1=tmpf;
  }
  c=poly[1]-u*b1-v*b2;
  d=poly[0]-v*b1;
  g=b2-u*f1-v*f2;
  h=b1-v*f1;
}

static int bairstow_iter(vector<ieee>& poly, ieee& u, ieee& v, ieee& error,int limit){
  //newton iteration
  ieee c,d,g,h;
  ieee det;
  ieee err2;

  int count=0;
  while(count<limit){
    bairstow_getp(poly,u,v,c,d,g,h);
    det=v*g*g+h*h-h*u*g;
    if(fabs(det)<1e-16){
      if((fabs(c)<1e-15)&&(fabs(d)<1e-15))break;//we hit the roots
      u+=AA/10;
      v+=PI/10;
      #if DEBUG>0
      fprintf(stderr,"bairstow solver: zero determinant occurred, retrying with u=%g, v=%g\n",(double)u,(double)v);
      #endif
      count++;
      continue;
    }
    det=1/det;
    ieee du=det*(-h*c+g*d);
    ieee dv=det*(-g*v*c+(g*u-h)*d);
    u-=du;
    v-=dv;
    error=sqrt(du*du+dv*dv);
    err2=sqrt(c*c+d*d);
    #if DEBUG>0
    fprintf(stderr,"bairstow solver: iteration step %d; error=%g, error(cd)=%g\n",count,(double)error,(double)err2);
    #endif
    if((error<1e-15)&&(err2<1e-9))break;
    count++;
  }
  if(err2>1e-9)error=1.;
  return count;
}

static int bairstow_step(vector<ieee>& poly, vector<ieee>& lower, array& out, int& addedzero){
  int n=poly.size()-1;
  #if DEBUG>0
  fprintf(stderr,"bairstow solver: starting step; n=%d\n",n);
  #endif
  //initial approximation
  ieee u=0;
  ieee v=-pow(fabs(poly[0]/poly[n]),2./n);

  #if DEBUG>0
  fprintf(stderr,"bairstow solver: initial approximation; u=%g, v=%g\n",(double)u,(double)v);
  #endif
  ieee error;
  int limit=n*24;
  int count=bairstow_iter(poly, u, v, error,limit);

  if((count==limit)&&(error>1e-9)){
    u=poly[n-1]/poly[n];
    v=poly[n-2]/poly[n];
    #if DEBUG>0
    fprintf(stderr,"bairstow solver: failed once, trying different approach; u=%g, v=%g\n",(double)u,(double)v);
    #endif
    count=bairstow_iter(poly, u, v, error,limit);
  }

  if((count==limit)&&(error>1e-9)&&(n%2==1)){
    #if DEBUG>0
    fprintf(stderr,"bairstow solver: failed twice, making polynomial of even order\n");
    #endif
    if(out.size()>0){
      ieee val=out.back().d;
      out.pop_back();
      //multiply poly by {-val 1}
      poly.push_back(poly[n]);
      for(int i=n;i>0;i--){
	poly[i]=poly[i-1]-poly[i]*val;
      }
      poly[0]*=-val;
    }else{
      poly.push_back(poly[n]);
      for(int i=n;i>0;i--){
	poly[i]=poly[i-1];
      }
      poly[0]=0;
      addedzero=1;
#if DEBUG>0
      fprintf(stderr,"bairstow solver: added artificial zero\n");
#endif
    }
    n++;
    u=0;
    v=-pow(fabs(poly[0]/poly[n]),2./n);
    #if DEBUG>0
    fprintf(stderr,"bairstow solver: new approximation; u=%g, v=%g\n",(double)u,(double)v);
    #endif
    count=bairstow_iter(poly, u, v, error,limit);
  }

  if(count==limit){
    if(error<1e-13){
    }else
    if(error<1e-11){
      fprintf(stderr,"bairstow solver: low accuracy (%g), possibly multiple root\n",(double)error);
    }else if(error<1e-7){
      fprintf(stderr,"bairstow solver: seriously low accuracy (%g), possibly multiple root\n",(double)error);
    }else{
      fprintf(stderr,"bairstow solver: failed to converge in %d steps\n",limit);
      return 1;
    }
  }
  //divide to obtain lower-order polynomial
  lower.resize(n-1);
  ieee b1=0;
  ieee b2=0;
  for(int i=n-2;i>=0;i--){
    ieee tmpb=poly[i+2]-u*b1-v*b2;
    lower[i]=tmpb;
    b2=b1;
    b1=tmpb;
  }
    #if DEBUG>0
    fprintf(stderr,"bairstow solver: successfully divided\n");
    #endif
  //found roots
  ieee disc=u*u-4*v;
  if(disc<-1e-14*u*u)return 0;
  if(disc<0)disc=0;
  out.push_back(value(((ieee)0.5)*(-u+sqrt(disc))));
  out.push_back(value(((ieee)0.5)*(-u-sqrt(disc))));
  return 0;
}

int bairstow(array& poly, array& out){
  int n2=poly.size();
  int n1=0;
  //trim zeroes
  for(int i=n2-1;i>=0;i--){
    if(poly[i].d==0){
      n2=i;
    }else break;
  }
  //front zeroes
  for(int i=0;i<n2;i++){
    if(poly[i].d==0){
      out.push_back(value((ieee)0));
      n1=i+1;
    }else{
      break;
    }
  }
  if(n2-n1>12){
    fprintf(stderr,"bairstow solver: warning, polynomial order is high (%d)\n",n2-n1-1);
  }
  switch(n2-n1){
    case 0:return 0;//all reals are solutions
    case 1:{//no zeroes
      return 0;
    }
    case 2:{
      out.push_back(value(-poly[n1].d/poly[n1+1].d));
      return 0;
    }
    case 3:{
      ieee u=poly[n1+1].d/poly[n1+2].d;
      ieee v=poly[n1].d/poly[n1+2].d;
      ieee disc=u*u-4*v;
      if(disc<-1e-14*u*u)return 0;
      if(disc<0)disc=0;
      out.push_back(value(((ieee)0.5)*(-u+sqrt(disc))));
      out.push_back(value(((ieee)0.5)*(-u-sqrt(disc))));
      return 0;
    }
    default:{
      int addedzero=0;
      vector<ieee> p;
      for(int i=n1;i<n2;i++){
	p.push_back(poly[i].d/poly[n2-1].d);
      }
      vector<ieee> p2;
      while(p.size()>3){
	if(bairstow_step(p,p2,out,addedzero)){
	  if(addedzero)out.clear();
	  return 1;
	}else{
	  p=p2;
	}
      }
      //now we got polynomial of order 1 or 2:
      if(p.size()==2){
	out.push_back(value(-p[0]/p[1]));
      }else{
	ieee u=p[1]/p[2];
	ieee v=p[0]/p[2];
	ieee disc=u*u-4*v;
	if(disc>-1e-14*u*u){
	  if(disc<0)disc=0;
	  out.push_back(value(((ieee)0.5)*(-u+sqrt(disc))));
	  out.push_back(value(((ieee)0.5)*(-u-sqrt(disc))));
	}
      }

      if(addedzero){
#if DEBUG>0
	fprintf(stderr,"bairstow solver: cleaning up added zero\n");
#endif
	int minindex=-1;
	ieee minz=1e-3;
	for(int j=0;j<out.size();j++){
	  ieee fbs=fabs(out[j].d);
	  if(fbs<minz){
	    minindex=j;
	    minz=fbs;
	  }
	}
	if(minindex>-1){
	  out.erase(out.begin()+minindex,out.begin()+minindex+1);
	}
	else{
	  fprintf(stderr,"bairstow solver: warning, added zero disappeared, one root may be wrong or missing\n");
	}
      }
      return 0;
    }
  }
}
